/** fMSX: portable MSX emulator ******************************/
/**                                                         **/
/**                         CommonMGL2.h                    **/
/**                                                         **/
/** This file contains standard screen refresh drivers      **/
/** common for X11, VGA, and other "chunky" bitmapped video **/
/** implementations. It also includes basic disk drivers    **/
/** and dummy sound drivers for fMSX.                       **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1994-1999                 **/
/**               John Stiles     1996                      **/
/**               Hiroyuki Yanai  2000                      **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/

static void  Sprites(byte Y);
static void  ColorSprites(byte Y,byte *ZBuf);
static void  RefreshBorder(byte Y,pixel C,int *PX,int *PY);
static void  ClearLine(byte Y,pixel C);

/** DiskPresent() *********************************************/
/** Return 1 if disk drive with a given ID is present.       **/
/**************************************************************/
byte DiskPresent(byte ID)
{ return((ID<2)&&(Drives[ID]>=0)); }

/** DiskRead() ************************************************/
/** Read requested sector from the drive into a buffer.      **/
/**************************************************************/
byte DiskRead(byte ID,byte *Buf,int N)
{
  if((ID<2)&&(Drives[ID]>=0))
    if(lseek(Drives[ID],N*512L,0)==N*512L)
      return(read(Drives[ID],Buf,512)==512);
  return(0);
}

/** DiskWrite() ***********************************************/
/** Write contents of the buffer into a given sector of the  **/
/** disk.                                                    **/
/**************************************************************/
byte DiskWrite(byte ID,byte *Buf,int N)
{
  if((ID<2)&&(Drives[ID]>=0))
    if(lseek(Drives[ID],N*512L,0)==N*512L)
      return(write(Drives[ID],Buf,512)==512);
  return(0);
}

/** RefreshScreen() ******************************************/
/** Refresh screen. This function is called in the end of   **/
/** refresh cycle to show the entire screen.                **/
/*************************************************************/
void RefreshScreen(void) { PutImage(); }

/** ClearLine() **********************************************/
/** Clear 256 pixels at line Y with color C.                   **/
/*************************************************************/
static void ClearLine(byte Y,register pixel C)
{
  int cur_width;

  set_color(C);

  cur_width = (is_width512()) ? (WIDTH512) : (WIDTH);
  draw_line(0, (int)Y, cur_width, (int)Y);
}

/** RefreshBorder() ******************************************/
/** This function is called from RefreshLine#() to refresh  **/
/** the screen border. It returns a pointer to the start of **/
/** scanline Y in XBuf.                                     **/
/*************************************************************/
void RefreshBorder(register byte Y,register pixel C,int *PX,int *PY)
#ifdef NARROW
{
  register int H;

  /* Set up the transparent color */
  XPal[0]=(!BGColor||SolidColor0)? XPal0:XPal[BGColor];

  /* Paint top of the border */
  set_color(C);
  if(!Y) {
    fill_rect(0, 0, WIDTH, 7);
  }

  /* Paint left/right borders */
  *PX=H=(int)(WIDTH-256)/2;
  *PY=(int)Y;
  draw_line(0, *PY, H - 1, *PY);
  draw_line(WIDTH-H, *PY, WIDTH, *PY);

  /* Paint bottom of the border */
  H=ScanLines212? 212:192;
  if(*PY==H-1) {
    fill_rect(0, H, WIDTH, (HEIGHT-H-7)-1);
  }
}
#else
{
  register int H;
  int cur_width;
  int hires_mode = 0;

  hires_mode = is_width512();
  cur_width = (hires_mode) ? (WIDTH512) : (WIDTH);

  /* Set up the transparent color */
  XPal[0]=(!BGColor||SolidColor0)? XPal0:XPal[BGColor];

  /* Paint top of the border */
  set_color(C);
  if(!Y) {
    fill_rect(0, 0, cur_width, 7);
  }

  /* Paint left/right borders */
  if(hires_mode) {
    *PX=0;
    *PY=(int)Y;
  } else {
    *PX=H=(int)(WIDTH-256)/2;
    *PY=(int)Y;
    draw_line(0, *PY, H - 1, *PY);
    draw_line(WIDTH-H, *PY, WIDTH, *PY);
  }

  /* Paint bottom of the border */
  H=ScanLines212? 212:192;
  if(*PY==H-1) {
    fill_rect(0, H, cur_width, (HEIGHT-H-7)-1);
  }
}
#endif /* NARROW */

/** Sprites() ************************************************/
/** This function is called from RefreshLine#() to refresh  **/
/** sprites in SCREENs 1-3.                                 **/
/*************************************************************/
void Sprites(register byte Y)
{
  register pixel C;
  register byte H,*PT,*AT;
  register unsigned int M;
  register int L,K;
  int PX,PY;
  int P[16];

  /* Assign initial values before counting */
  H=Sprites16x16? 16:8;
  C=0;M=0;L=0;
  AT=SprTab-4;

  /* Count displayed sprites */
  do
  {
    M<<=1;AT+=4;L++;    /* Iterating through SprTab      */
    K=AT[0];            /* K = sprite Y coordinate       */
    if(K==208) break;   /* Iteration terminates if Y=208 */
    if(K>256-H) K-=256; /* Y coordinate may be negative  */

    /* Mark all valid sprites with 1s, break at MAXSPRITE1 sprites */
    if((Y>K)&&(Y<=K+H)) { M|=1;C++;if(C==MAXSPRITE1) break; }
  }
  while(L<32);

  /* Draw all marked sprites */
  for(;M;M>>=1,AT-=4)
    if(M&1)
    {
      C=AT[3];                  /* C = sprite attributes */
      L=C&0x80? AT[1]-32:AT[1]; /* Sprite may be shifted left by 32 */
      C&=0x0F;                  /* C = sprite color */

      if((L<256)&&(L>-H)&&C)
      {
        K=AT[0];                /* K = sprite Y coordinate */
        if(K>256-H) K-=256;     /* Y coordinate may be negative */

        PX=(WIDTH-256)/2+L;
        PY=(int)Y;

        PT=SprGen+((int)(H>8? AT[2]&0xFC:AT[2])<<3)+Y-K-1;
        C=XPal[C];

        /* Mask 1: clip left sprite boundary */
        K=L>=0? 0x0FFFF:(0x10000>>-L)-1;
        /* Mask 2: clip right sprite boundary */
        if(L>256-H) K^=((0x00200>>(H-8))<<(L-257+H))-1;
        /* Get and clip the sprite data */
        K&=((int)PT[0]<<8)|(H>8? PT[16]:0x00);

        get_pixstream(PX,PY,P,16,DIR_NORTH,0);

        /* Draw left 8 pixels of the sprite */
        if(K&0xFF00)
        {
          if(K&0x8000) P[0]=C; if(K&0x4000) P[1]=C;
          if(K&0x2000) P[2]=C; if(K&0x1000) P[3]=C;
          if(K&0x0800) P[4]=C; if(K&0x0400) P[5]=C;
          if(K&0x0200) P[6]=C; if(K&0x0100) P[7]=C;
        }

        /* Draw right 8 pixels of the sprite */
        if(K&0x00FF)
        {
          if(K&0x0080) P[8]=C;  if(K&0x0040) P[9]=C;
          if(K&0x0020) P[10]=C; if(K&0x0010) P[11]=C;
          if(K&0x0008) P[12]=C; if(K&0x0004) P[13]=C;
          if(K&0x0002) P[14]=C; if(K&0x0001) P[15]=C;
        }

        put_pixstream(PX,PY,P,16,DIR_NORTH);
      }
    }
}

/** ColorSprites() *******************************************/
/** This function is called from RefreshLine#() to refresh  **/
/** color sprites in SCREENs 4-8. The result is returned in **/
/** ZBuf, whose size must be 304 bytes (32+256+16).         **/
/*************************************************************/
void ColorSprites(register byte Y,byte *ZBuf)
{
  register byte C,H,J,OrThem;
  register byte *P,*PT,*AT;
  register int L,K;
  register unsigned int M;

  /* Clear ZBuffer and exit if sprites are off */
  memset(ZBuf+32,0,256);
  if(SpritesOFF) return;

  /* Assign initial values before counting */
  H=Sprites16x16? 16:8;
  C=0;M=0;L=0;
  AT=SprTab-4;
  OrThem=0x00;

  /* Count displayed sprites */
  do
  {
    M<<=1;AT+=4;L++;          /* Iterating through SprTab      */
    K=AT[0];                  /* Read Y from SprTab            */
    if(K==216) break;         /* Iteration terminates if Y=216 */
    K=(byte)(K-VDP[23]);      /* Sprite's actual Y coordinate  */
    if(K>256-H) K-=256;       /* Y coordinate may be negative  */

    /* Mark all valid sprites with 1s, break at MAXSPRITE2 sprites */
    if((Y>K)&&(Y<=K+H)) { M|=1;C++;if(C==MAXSPRITE2) break; }
  }
  while(L<32);

  /* Draw all marked sprites */
  for(;M;M>>=1,AT-=4)
    if(M&1)
    {
      K=(byte)(AT[0]-VDP[23]); /* K = sprite Y coordinate */
      if(K>256-H) K-=256;      /* Y coordinate may be negative */

      J=Y-K-1;
      C=SprTab[-0x0200+((AT-SprTab)<<2)+J];
      OrThem|=C&0x40;

      if(C&0x0F)
      {
        PT=SprGen+((int)(H>8? AT[2]&0xFC:AT[2])<<3)+J;
        P=ZBuf+AT[1]+(C&0x80? 0:32);
        C&=0x0F;
        J=PT[0];

        if(OrThem&0x20)
        {
          if(J&0x80) P[0]|=C;if(J&0x40) P[1]|=C;
          if(J&0x20) P[2]|=C;if(J&0x10) P[3]|=C;
          if(J&0x08) P[4]|=C;if(J&0x04) P[5]|=C;
          if(J&0x02) P[6]|=C;if(J&0x01) P[7]|=C;
          if(H>8)
          {
            J=PT[16];
            if(J&0x80) P[8]|=C; if(J&0x40) P[9]|=C;
            if(J&0x20) P[10]|=C;if(J&0x10) P[11]|=C;
            if(J&0x08) P[12]|=C;if(J&0x04) P[13]|=C;
            if(J&0x02) P[14]|=C;if(J&0x01) P[15]|=C;
          }
        }
        else
        {
          if(J&0x80) P[0]=C;if(J&0x40) P[1]=C;
          if(J&0x20) P[2]=C;if(J&0x10) P[3]=C;
          if(J&0x08) P[4]=C;if(J&0x04) P[5]=C;
          if(J&0x02) P[6]=C;if(J&0x01) P[7]=C;
          if(H>8)
          {
            J=PT[16];
            if(J&0x80) P[8]=C; if(J&0x40) P[9]=C;
            if(J&0x20) P[10]=C;if(J&0x10) P[11]=C;
            if(J&0x08) P[12]=C;if(J&0x04) P[13]=C;
            if(J&0x02) P[14]=C;if(J&0x01) P[15]=C;
          }
        }
      }

      /* Update overlapping flag */
      OrThem>>=1;
    }
}

/** RefreshLineF() *******************************************/
/** Dummy refresh function called for non-existing screens. **/
/*************************************************************/
void RefreshLineF(register byte Y)
{
  int PX, PY;

  if(Verbose>1)
    printf
    (
      "ScrMODE %d: ChrTab=%X ChrGen=%X ColTab=%X SprTab=%X SprGen=%X\n",
      ScrMode,ChrTab-VRAM,ChrGen-VRAM,ColTab-VRAM,SprTab-VRAM,SprGen-VRAM
    );

  RefreshBorder(Y,XPal[BGColor],&PX,&PY);
  ClearLine(Y,XPal[BGColor]);
}

/** RefreshLine0() *******************************************/
/** Refresh line Y (0..191/211) of SCREEN0.                 **/
/*************************************************************/
void RefreshLine0(register byte Y)
{
  register pixel FC,BC;
  register byte X,*T,*G;
  int PX,PY;
  int P[10];

  BC=XPal[BGColor];
  RefreshBorder(Y,BC,&PX,&PY);

  if(!ScreenON) ClearLine(Y,BC);
  else
  {
    set_color(BC);
    draw_line(PX,PY,PX+7,PY);

    G=(UseFont&&FontBuf? FontBuf:ChrGen)+(Y&0x07);
    T=ChrTab+40*(Y>>3);
    FC=XPal[FGColor];
    PX+=8;

    for(X=0;X<40;X++,T++,PX+=6)
    {
      Y=G[(int)*T<<3];
      P[0]=Y&0x80? FC:BC; P[1]=Y&0x40? FC:BC;
      P[2]=Y&0x20? FC:BC; P[3]=Y&0x10? FC:BC;
      P[4]=Y&0x08? FC:BC; P[5]=Y&0x04? FC:BC;

      P[6]=BC; P[7]=BC;
      put_pixstream(PX,PY,P,8,DIR_NORTH);
    }

    PX++;
    set_color(BC);
    draw_line(PX,PY,PX+7,PY);
  }
}

/** RefreshLine1() *******************************************/
/** Refresh line Y (0..191/211) of SCREEN1, including       **/
/** sprites in this line.                                   **/
/*************************************************************/
void RefreshLine1(register byte Y)
{
  register pixel FC,BC;
  register byte K,X,*T,*G;
  int PX,PY;
  int P[8];

  RefreshBorder(Y,XPal[BGColor],&PX,&PY);

  if(!ScreenON) ClearLine(Y,XPal[BGColor]);
  else
  {
    G=(UseFont&&FontBuf? FontBuf:ChrGen)+(Y&0x07);
    T=ChrTab+((int)(Y&0xF8)<<2);

    for(X=0;X<32;X++,T++,PX+=8)
    {
      K=ColTab[*T>>3];
      FC=XPal[K>>4];
      BC=XPal[K&0x0F];
      K=G[(int)*T<<3];

      P[0]=K&0x80? FC:BC; P[1]=K&0x40? FC:BC;
      P[2]=K&0x20? FC:BC; P[3]=K&0x10? FC:BC;
      P[4]=K&0x08? FC:BC; P[5]=K&0x04? FC:BC;
      P[6]=K&0x02? FC:BC; P[7]=K&0x01? FC:BC;
      put_pixstream(PX,PY,P,8,DIR_NORTH);
    }

    if(!SpritesOFF) Sprites(Y);
  }
}

/** RefreshLine2() *******************************************/
/** Refresh line Y (0..191/211) of SCREEN2, including       **/
/** sprites in this line.                                   **/
/*************************************************************/
void RefreshLine2(register byte Y)
{
  register pixel FC,BC;
  register byte K,X,*T,*PGT,*CLT;
  register int J;
  int PX,PY;
  int P[8];

  RefreshBorder(Y,XPal[BGColor],&PX,&PY);

  if(!ScreenON) ClearLine(Y,XPal[BGColor]);
  else
  {
    T=ChrTab+((int)(Y&0xF8)<<2);
    J=((int)(Y&0xC0)<<5)+(Y&0x07);
    PGT=ChrGen+J;
    CLT=ColTab+J;

    for(X=0;X<32;X++,T++,PX+=8)
    {
      J=(int)*T<<3;
      K=CLT[J];
      FC=XPal[K>>4];
      BC=XPal[K&0x0F];
      K=PGT[J];

      P[0]=K&0x80? FC:BC; P[1]=K&0x40? FC:BC;
      P[2]=K&0x20? FC:BC; P[3]=K&0x10? FC:BC;
      P[4]=K&0x08? FC:BC; P[5]=K&0x04? FC:BC;
      P[6]=K&0x02? FC:BC; P[7]=K&0x01? FC:BC;
      put_pixstream(PX,PY,P,8,DIR_NORTH);
    }

    if(!SpritesOFF) Sprites(Y);
  }
}

/** RefreshLine3() *******************************************/
/** Refresh line Y (0..191/211) of SCREEN3, including       **/
/** sprites in this line.                                   **/
/*************************************************************/
void RefreshLine3(register byte Y)
{
  register byte X,K,*T,*G;
  int PX,PY;

  RefreshBorder(Y,XPal[BGColor],&PX,&PY);

  if(!ScreenON) ClearLine(Y,XPal[BGColor]);
  else
  {
    T=ChrTab+((int)(Y&0xF8)<<2);
    G=ChrGen+((Y&0x1C)>>2);

    for(X=0;X<32;X++,T++,PX+=8)
    {
      K=G[(int)*T<<3];
      set_color(XPal[K>>4]);
      draw_line(PX, PY, PX + 3, PY);

      set_color(XPal[K&0x0F]);
      draw_line(PX+4,PY,PX+7,PY);
    }
  }
}

/** RefreshLine4() *******************************************/
/** Refresh line Y (0..191/211) of SCREEN4, including       **/
/** sprites in this line.                                   **/
/*************************************************************/
void RefreshLine4(register byte Y)
{
  register pixel FC,BC;
  register byte I,X,C,*T,*R;
  register int J;
  byte ZBuf[304];
  int PX,PY;
  int P[8];

  RefreshBorder(Y,XPal[BGColor],&PX,&PY);

  if(!ScreenON) ClearLine(Y,XPal[BGColor]);
  else
  {
    ColorSprites(Y,ZBuf);
    R=ZBuf+32;
    Y+=VDP[23];
    T=ChrTab+((int)(Y&0xF8)<<2);

    for(X=0;X<32;X++,R+=8,PX+=8,T++)
    {
      J=((int)(Y&0xC0)<<5)+((int)*T<<3)+(Y&0x07);
      I=ColTab[J];
      FC=XPal[I>>4];
      BC=XPal[I&0x0F];
      I=ChrGen[J];

      C=R[0]; P[0]=C? XPal[C]:(I&0x80)? FC:BC;
      C=R[1]; P[1]=C? XPal[C]:(I&0x40)? FC:BC;
      C=R[2]; P[2]=C? XPal[C]:(I&0x20)? FC:BC;
      C=R[3]; P[3]=C? XPal[C]:(I&0x10)? FC:BC;
      C=R[4]; P[4]=C? XPal[C]:(I&0x08)? FC:BC;
      C=R[5]; P[5]=C? XPal[C]:(I&0x04)? FC:BC;
      C=R[6]; P[6]=C? XPal[C]:(I&0x02)? FC:BC;
      C=R[7]; P[7]=C? XPal[C]:(I&0x01)? FC:BC;
      put_pixstream(PX,PY,P,8,DIR_NORTH);
    }
  }
}

/** RefreshLine5() *******************************************/
/** Refresh line Y (0..191/211) of SCREEN5, including       **/
/** sprites in this line.                                   **/
/*************************************************************/
void RefreshLine5(register byte Y)
{
  register byte I,X,*T,*R;
  byte ZBuf[304];
  int PX,PY;
  int P[16];

  RefreshBorder(Y,XPal[BGColor],&PX,&PY);

  if(!ScreenON) ClearLine(Y,XPal[BGColor]);
  else
  {
    ColorSprites(Y,ZBuf);
    R=ZBuf+32;
    Y+=VDP[23];
    T=ChrTab+((int)Y<<7);

    for(X=0;X<16;X++,R+=16,PX+=16,T+=8)
    {
      I=R[0];  P[0]=XPal[I? I:T[0]>>4];
      I=R[1];  P[1]=XPal[I? I:T[0]&0x0F];
      I=R[2];  P[2]=XPal[I? I:T[1]>>4];
      I=R[3];  P[3]=XPal[I? I:T[1]&0x0F];
      I=R[4];  P[4]=XPal[I? I:T[2]>>4];
      I=R[5];  P[5]=XPal[I? I:T[2]&0x0F];
      I=R[6];  P[6]=XPal[I? I:T[3]>>4];
      I=R[7];  P[7]=XPal[I? I:T[3]&0x0F];
      I=R[8];  P[8]=XPal[I? I:T[4]>>4];
      I=R[9];  P[9]=XPal[I? I:T[4]&0x0F];
      I=R[10]; P[10]=XPal[I? I:T[5]>>4];
      I=R[11]; P[11]=XPal[I? I:T[5]&0x0F];
      I=R[12]; P[12]=XPal[I? I:T[6]>>4];
      I=R[13]; P[13]=XPal[I? I:T[6]&0x0F];
      I=R[14]; P[14]=XPal[I? I:T[7]>>4];
      I=R[15]; P[15]=XPal[I? I:T[7]&0x0F];
      put_pixstream(PX,PY,P,16,DIR_NORTH);
    }
  }
}

/** RefreshScr8() ********************************************/
/** Function to be called to update SCREEN 8.               **/
/*************************************************************/
void RefreshLine8(register byte Y)
{
  static byte SprToScr[16] =
  {
    0x00,0x02,0x10,0x12,0x80,0x82,0x90,0x92,
    0x49,0x4B,0x59,0x5B,0xC9,0xCB,0xD9,0xDB
  };
  register byte C,X,*T,*R;
  byte ZBuf[304];
  int PX,PY;
  int P[2];

  RefreshBorder(Y,XPal[BGColor],&PX,&PY);

  if(!ScreenON) ClearLine(Y,XPal[BGColor]);
  else
  {
    ColorSprites(Y,ZBuf);
    R=ZBuf+32;
    Y+=VDP[23];
    T=ChrTab+((int)Y<<8);

    for(X=0;X<128;X++,T+=2,R+=2,PX+=2)
    {
      C=R[0]; P[0]=BPal[C? SprToScr[C]:T[0]];
      C=R[1]; P[1]=BPal[C? SprToScr[C]:T[1]];
      put_pixstream(PX,PY,P,2,DIR_NORTH);
    }
  }
}

#ifdef NARROW

/** RefreshScr6() ********************************************/
/** Function to be called to update SCREEN 6.               **/
/*************************************************************/
void RefreshLine6(register byte Y)
{
  register byte X,*T,*R,C;
  byte ZBuf[304];
  int PX,PY;
  int P[8];

  RefreshBorder(Y,XPal[BGColor&0x03],&PX,&PY);

  if(!ScreenON) ClearLine(Y,XPal[BGColor&0x03]);
  else
  {
    ColorSprites(Y,ZBuf);
    R=ZBuf+32;
    T=ChrTab+(((int)Y+VDP[23])<<7);

    for(X=0;X<32;X++)
    {
      C=R[0]; P[0]=XPal[C? C:T[0]>>6];
      C=R[1]; P[1]=XPal[C? C:(T[0]>>2)&0x03];
      C=R[2]; P[2]=XPal[C? C:T[1]>>6];
      C=R[3]; P[3]=XPal[C? C:(T[1]>>2)&0x03];
      C=R[4]; P[4]=XPal[C? C:T[2]>>6];
      C=R[5]; P[5]=XPal[C? C:(T[2]>>2)&0x03];
      C=R[6]; P[6]=XPal[C? C:T[3]>>6];
      C=R[7]; P[7]=XPal[C? C:(T[3]>>2)&0x03];
      put_pixstream(PX,PY,P,8,DIR_NORTH);
      R+=8;PX+=8;T+=4;
    }
  }
}
  
/** RefreshScr7() ********************************************/
/** Function to be called to update SCREEN 7.               **/
/*************************************************************/
void RefreshLine7(register byte Y)
{
  register byte C,X,*T,*R;
  byte ZBuf[304];
  int PX,PY;
  int P[8];

  RefreshBorder(Y,XPal[BGColor],&PX,&PY);

  if(!ScreenON) ClearLine(Y,XPal[BGColor]);
  else
  {
    ColorSprites(Y,ZBuf);
    R=ZBuf+32;
    T=ChrTab+(((int)Y+VDP[23])<<8);

    for(X=0;X<32;X++)
    {
      C=R[0]; P[0]=XPal[C? C:T[0]>>4];
      C=R[1]; P[1]=XPal[C? C:T[1]>>4];
      C=R[2]; P[2]=XPal[C? C:T[2]>>4];
      C=R[3]; P[3]=XPal[C? C:T[3]>>4];
      C=R[4]; P[4]=XPal[C? C:T[4]>>4];
      C=R[5]; P[5]=XPal[C? C:T[5]>>4];
      C=R[6]; P[6]=XPal[C? C:T[6]>>4];
      C=R[7]; P[7]=XPal[C? C:T[7]>>4];
      put_pixstream(PX,PY,P,8,DIR_NORTH);
      R+=8;PX+=8;T+=8;
    }
  }
}

/** RefreshTx80() ********************************************/
/** Function to be called to update TEXT80.                 **/
/*************************************************************/
void RefreshLineTx80(register byte Y)
{
  register pixel FC,BC;
  register byte X,M,*T,*C,*G;
  int PX,PY;
  int P[4];

  BC=XPal[BGColor];
  RefreshBorder(Y,BC,&PX,&PY);

  if(!ScreenON) ClearLine(Y,BC);
  else
  {
    G=(UseFont&&FontBuf? FontBuf:ChrGen)+(Y&0x07);
    T=ChrTab+80*(Y>>3);
    C=ColTab+10*(Y>>3);
    PX-=32;
    M=0x80;

    for(X=0;X<80;X++,T++,PX+=4)
    {
      Y=*(G+((int)*T<<3));
      if(*C&M) { FC=XPal[XFGColor];BC=XPal[XBGColor]; }
      else     { FC=XPal[FGColor];BC=XPal[BGColor]; }
      P[0]=Y&0xC0? FC:BC;
      P[1]=Y&0x30? FC:BC;
      P[2]=Y&0x0C? FC:BC;
      P[3]=BC;
      put_pixstream(PX,PY,P,4,DIR_NORTH);
      if(!(M>>=1)) { M=0x80;C++; } 
    }
  }
}

#else  /* !NARROW */

/** RefreshScr6() ********************************************/
/** Function to be called to update SCREEN 6.               **/
/*************************************************************/
void RefreshLine6(register byte Y)
{
  register byte X,*T,*R,C;
  byte ZBuf[304];
  int PX,PY;
  int P[16];

  RefreshBorder(Y,XPal[BGColor&0x03],&PX,&PY);

  if(!ScreenON) ClearLine(Y,XPal[BGColor&0x03]);
  else
  {
    ColorSprites(Y,ZBuf);
    R=ZBuf+32;
    T=ChrTab+(((int)Y+VDP[23])<<7);

    for(X=0;X<32;X++)
    {
      C=R[0]; P[0] =XPal[C? C:T[0]>>6];
              P[1] =XPal[C? C:(T[0]>>4)&0x03];
      C=R[1]; P[2] =XPal[C? C:(T[0]>>2)&0x03];
              P[3] =XPal[C? C:T[0]&0x03];  

      C=R[2]; P[4] =XPal[C? C:T[1]>>6];
              P[5] =XPal[C? C:(T[1]>>4)&0x03];
      C=R[3]; P[6] =XPal[C? C:(T[1]>>2)&0x03];
              P[7] =XPal[C? C:T[1]&0x03];

      C=R[4]; P[8] =XPal[C? C:T[2]>>6];
              P[9] =XPal[C? C:(T[2]>>4)&0x03];
      C=R[5]; P[10]=XPal[C? C:(T[2]>>2)&0x03];
              P[11]=XPal[C? C:T[2]&0x03];

      C=R[6]; P[12]=XPal[C? C:T[3]>>6];
              P[13]=XPal[C? C:(T[3]>>4)&0x03];
      C=R[7]; P[14]=XPal[C? C:(T[3]>>2)&0x03];
              P[15]=XPal[C? C:T[3]&0x03];

      put_pixstream(PX,PY,P,16,DIR_NORTH);
      R+=8;PX+=16;T+=4;
    }
  }
}
  
/** RefreshScr7() ********************************************/
/** Function to be called to update SCREEN 7.               **/
/*************************************************************/
void RefreshLine7(register byte Y)
{
  register byte C,X,*T,*R;
  byte ZBuf[304];
  int PX,PY;
  int P[16];

  RefreshBorder(Y,XPal[BGColor],&PX,&PY);

  if(!ScreenON) ClearLine(Y,XPal[BGColor]);
  else
  {
    ColorSprites(Y,ZBuf);
    R=ZBuf+32;
    T=ChrTab+(((int)Y+VDP[23])<<8);

    for(X=0;X<32;X++)
    {
      C=R[0]; P[0] =XPal[C? C:T[0]>>4];
              P[1] =XPal[C? C:T[0]&0x0F];
      C=R[1]; P[2] =XPal[C? C:T[1]>>4];
              P[3] =XPal[C? C:T[1]&0x0F];
      C=R[2]; P[4] =XPal[C? C:T[2]>>4];
              P[5] =XPal[C? C:T[2]&0x0F];
      C=R[3]; P[6] =XPal[C? C:T[3]>>4];
              P[7] =XPal[C? C:T[3]&0x0F];
      C=R[4]; P[8] =XPal[C? C:T[4]>>4];
              P[9] =XPal[C? C:T[4]&0x0F];
      C=R[5]; P[10]=XPal[C? C:T[5]>>4];
              P[11]=XPal[C? C:T[5]&0x0F];
      C=R[6]; P[12]=XPal[C? C:T[6]>>4];
              P[13]=XPal[C? C:T[6]&0x0F];
      C=R[7]; P[14]=XPal[C? C:T[7]>>4];
              P[15]=XPal[C? C:T[7]&0x0F];
      put_pixstream(PX,PY,P,16,DIR_NORTH);
      R+=8;PX+=16;T+=8;
    }
  }
}

/** RefreshTx80() ********************************************/
/** Function to be called to update TEXT80.                 **/
/*************************************************************/
void RefreshLineTx80(register byte Y)
{
  register pixel FC,BC;
  register byte X,M,*T,*C,*G;
  int PX,PY;
  int P[8];

  BC=XPal[BGColor];
  RefreshBorder(Y,BC,&PX,&PY);

  if(!ScreenON) ClearLine(Y,BC);
  else
  {
    G=(UseFont&&FontBuf? FontBuf:ChrGen)+(Y&0x07);
    T=ChrTab+80*(Y>>3);
    C=ColTab+10*(Y>>3);
    //PX-=32;
    M=0x80;

    for(X=0;X<80;X++,T++,PX+=6)
    {
      Y=*(G+((int)*T<<3));
      if(*C&M) { FC=XPal[XFGColor];BC=XPal[XBGColor]; }
      else     { FC=XPal[FGColor];BC=XPal[BGColor]; }

      P[0]=Y&0x80? FC:BC;
      P[1]=Y&0x40? FC:BC;
      P[2]=Y&0x20? FC:BC;
      P[3]=Y&0x10? FC:BC;

      P[4]=Y&0x08? FC:BC;
      P[5]=Y&0x04? FC:BC;
      P[6]=P[7]=BC;
      put_pixstream(PX,PY,P,8,DIR_NORTH);
      if(!(M>>=1)) { M=0x80;C++; } 
    }
  }
}

#endif /* NARROW */
