/** fMSX: portable MSX emulator ******************************/
/**                                                         **/
/**                          MSX.c                          **/
/**                                                         **/
/** This file contains implementation for the MSX-specific  **/
/** hardware: slots, memory mapper, PPIs, VDP, PSG, clock,  **/
/** etc. Initialization code and definitions needed for the **/
/** machine-dependent drivers are also here.                **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1994-1999                 **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/

#include "MSX.h"
#include "V9938.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <time.h>

#ifdef ZLIB
#include <zlib.h>
#endif

#ifndef O_BINARY
#define O_BINARY 0
#endif

#define PRINTOK        if(Verbose) puts("OK")
#define PRINTFAILED    if(Verbose) puts("FAILED")
#define PRINTRESULT(R) if(Verbose) puts(R? "OK":"FAILED")

/** User-defined parameters for fMSX *************************/
byte Verbose     = 1;              /* Debug msgs ON/OFF      */
byte UPeriod     = 2;              /* Interrupts/scr. update */
int  VPeriod     = 75000;          /* CPU cycles per VBlank  */
int  HPeriod     = 250;            /* CPU cycles per HBlank  */
byte SaveCMOS    = 0;              /* Save CMOS.ROM on exit  */
byte MSXVersion  = 1;              /* 0=MSX1,1=MSX2,2=MSX2+  */
byte JoyTypeA    = 0;              /* 0=None,1=Jstk,2=Mouse  */
byte JoyTypeB    = 0;
byte ROMTypeA    = 0;              /* MegaROM types          */
byte ROMTypeB    = 0;
int  RAMPages    = 4;              /* Number of RAM pages    */
int  VRAMPages   = 2;              /* Number of VRAM pages   */

byte ExitNow;                      /* 1: Exit the emulator   */

/** Main hardware: CPU, RAM, VRAM, mappers *******************/
Z80 CPU;                           /* Z80 CPU state and regs */

byte *VRAM,*VPAGE;                 /* Video RAM              */

byte *RAM[8];                      /* Main RAM [8x8kB pages] */
byte *EmptyRAM;                    /* Empty RAM page [8kB]   */
byte *MemMap[4][4][8];   /* Memory maps [PPage][SPage][Addr] */

byte *RAMMap[256];                 /* Blocks for RAM Mapper  */
byte RAMMapper[4];                 /* RAM Mapper state       */
byte RAMMask;                      /* RAM Mapper mask        */

byte *ROMMap[2][256];              /* Blocks for ROM Mappers */
byte ROMMapper[2][4];              /* ROM Mappers state      */
byte ROMMask[2];                   /* ROM Mapper masks       */

byte EnWrite[4];                   /* 1 if write enabled     */
byte PSL[4],SSL[4];                /* Lists of current slots */
byte PSLReg,SSLReg;      /* Storage for A8h port and (FFFFh) */

/** Memory blocks to free in TrashMSX() **********************/
byte *Chunks[256];                 /* Memory blocks to free  */
byte CCount;                       /* Number of memory blcks */

/** Cartridge files used by fMSX *****************************/
char *CartA      = "CARTA.ROM";    /* Cartridge A ROM file   */
char *CartB      = "CARTB.ROM";    /* Cartridge B ROM file   */

/** Disk images used by fMSX *********************************/
char *DiskA      = "DRIVEA.DSK";   /* Drive A disk image     */
char *DiskB      = "DRIVEB.DSK";   /* Drive B disk image     */
int  Drives[2]   = { -1,-1 };      /* Disk image files       */

/** Fixed font used by fMSX **********************************/
char *FontName   = "DEFAULT.FNT";  /* Font file for text     */
byte *FontBuf;                     /* Font for text modes    */
byte UseFont     = 0;              /* Use ext. font when 1   */

/** Soundtrack logging ***************************************/
char *SndName    = NULL;           /* Sound log file         */
FILE *SndStream;

/** Printer **************************************************/
char *PrnName    = NULL;           /* Printer redirect. file */
FILE *PrnStream;

/** Cassette tape ********************************************/
char *CasName    = "DEFAULT.CAS";  /* Tape image file        */
FILE *CasStream;

/** Serial port **********************************************/
char *ComName    = NULL;           /* Serial redirect. file  */
FILE *ComIStream;
FILE *ComOStream;

/** Kanji font ROM *******************************************/
byte *Kanji;                       /* Kanji ROM 4096x32      */
int  KanLetter;                    /* Current letter index   */
byte KanCount;                     /* Byte count 0..31       */

/** Keyboard and mouse ***************************************/
byte KeyMap[16];                   /* Keyboard map           */
byte Buttons[2];                   /* Mouse button states    */
byte MouseDX[2],MouseDY[2];        /* Mouse offsets          */
byte OldMouseX[2],OldMouseY[2];    /* Old mouse coordinates  */
byte MCount[2];                    /* Mouse nibble counter   */

/** General I/O registers ************************************/
byte IOReg;                        /* Storage for AAh port   */
byte PPIReg;                       /* PPI control register   */
byte COMReg;                       /* 8251 ACIA control reg. */

/** Sound hardware: PSG, SCC, OPLL ***************************/
byte OPLLReg;                      /* OPLL register number   */
byte PSGReg;                       /* PSG register number    */
byte OPLL[256];                    /* OPLL registers         */
byte PSG[16];                      /* PSG registers          */
byte SCC[256];                     /* SCC registers          */

/** Real-time clock ******************************************/
byte RTCReg,RTCMode;               /* RTC register numbers   */
byte RTC[4][13];                   /* RTC registers          */

/** Video processor ******************************************/
byte *ChrGen,*ChrTab,*ColTab;      /* VDP tables [screens]   */
byte *SprGen,*SprTab,*SprCol;      /* VDP tables [sprites]   */
word VAddr;                        /* VRAM address in VDP    */
byte VKey,PKey,WKey;               /* Status keys for VDP    */
byte FGColor,BGColor;              /* Colors                 */
byte XFGColor,XBGColor;            /* Second set of colors   */
byte ScrMode;                      /* Current screen mode    */
byte VDP[64],VDPStatus[16];        /* VDP registers          */
byte ScanLine;                     /* Current scanline       */
byte VDPData;                      /* VDP data buffer        */

/** Places in DiskROM to be patched with ED FE C9 ************/
word DiskPatches[] = { 0x4010,0x4013,0x4016,0x401C,0x401F,0 };

/** Places in BIOS to be patched with ED FE C9 ***************/
word BIOSPatches[] = { 0x00E1,0x00E4,0x00E7,0x00EA,0x00ED,0x00F0,0x00F3,0 };

/** Screen Mode Handlers [number of screens + 1] *************/
void (*RefreshLine[MAXSCREEN+2])(byte Y) =
{
  RefreshLine0,   /* SCR 0:TEXT 40x24  */
  RefreshLine1,   /* SCR 1:TEXT 32x24  */
  RefreshLine2,   /* SCR 2:BLK 256x192 */
  RefreshLine3,   /* SCR 3:64x48x16    */
  RefreshLine4,   /* SCR 4:BLK 256x192 */
  RefreshLine5,   /* SCR 5:256x192x16  */
  RefreshLine6,   /* SCR 6:512x192x4   */
  RefreshLine7,   /* SCR 7:512x192x16  */
  RefreshLine8,   /* SCR 8:256x192x256 */
  RefreshLineTx80 /* SCR 0:TEXT 80x24  */
};

/** VDP Address Register Masks *******************************/
struct { byte R2,R3,R4,R5; } MSK[MAXSCREEN+2] =
{
  { 0x7F,0x00,0x3F,0x00 }, /* SCR 0:TEXT 40x24  */
  { 0x7F,0xFF,0x3F,0xFF }, /* SCR 1:TEXT 32x24  */
  { 0x7F,0x80,0x3C,0xFF }, /* SCR 2:BLK 256x192 */
  { 0x7F,0x00,0x3F,0xFF }, /* SCR 3:64x48x16    */
  { 0x7F,0x80,0x3C,0xFC }, /* SCR 4:BLK 256x192 */
  { 0x60,0x00,0x00,0xFC }, /* SCR 5:256x192x16  */
  { 0x60,0x00,0x00,0xFC }, /* SCR 6:512x192x4   */
  { 0x20,0x00,0x00,0xFC }, /* SCR 7:512x192x16  */
  { 0x20,0x00,0x00,0xFC }, /* SCR 8:256x192x256 */
  { 0x7C,0xF8,0x3F,0x00 }  /* SCR 0:TEXT 80x24  */
};

/** Internal Functions ***************************************/
/** These functions are defined and internally used by the  **/
/** code in MSX.c.                                          **/
/*************************************************************/
byte *LoadROM(char *Name,unsigned int Size,byte *Buf);
int  LoadCart(char *Name,int Slot);
void MapROM(word Addr,byte Value);/* Switch MegaROM banks            */
void SSlot(byte Value);           /* Switch secondary slots          */
void VDPOut(byte Reg,byte Value); /* Write value into a VDP register */
void Printer(byte V);             /* Send a character to a printer   */
void SIOOut(byte V);              /* Send a character to serial port */
byte SIOIn(void);                 /* Get character from serial port  */
void CheckSprites(void);          /* Check collisions and 5th sprite */
byte RTCIn(byte R);               /* Read RTC registers              */

/** StartMSX() ***********************************************/
/** Allocate memory, load ROM images, initialize hardware,  **/
/** CPU and start the emulation. This function returns 0 in **/
/** the case of failure.                                    **/
/*************************************************************/
int StartMSX(void)
{
  /*** MSX versions: ***/
  static char *Versions[] = { "MSX","MSX2","MSX2+" };
  /*** Joystick types: ***/
  static char *JoyTypes[] =
  {
    "nothing","normal joystick",
    "mouse in joystick mode","mouse in real mode"
  };
  /*** CMOS ROM default values: ***/
  static byte RTCInit[4][13]  =
  {
    {  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
    {  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
    {  0, 0, 0, 0,40,80,15, 4, 4, 0, 0, 0, 0 },
    {  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }
  };
  /*** PSG initial register states: ***/
  static byte PSGInit[16]  = { 0,0,0,0,0,0,0,0xFD,0,0,0,0,0,0,0xFF,0 };
  /*** VDP status register states: ***/
  static byte VDPSInit[16] = { 0x9F,0,0x4C,0,0,0,0,0,0,0,0,0,0,0,0,0 };
  /*** VDP control register states: ***/
  static byte VDPInit[64]  =
  {
    0x00,0x10,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00
  };
  /*** Initial palette: ***/
  static byte PalInit[16][3] =
  {
    {0x00,0x00,0x00},{0x00,0x00,0x00},{0x20,0xC0,0x20},{0x60,0xE0,0x60},
    {0x20,0x20,0xE0},{0x40,0x60,0xE0},{0xA0,0x20,0x20},{0x40,0xC0,0xE0},
    {0xE0,0x20,0x20},{0xE0,0x60,0x60},{0xC0,0xC0,0x20},{0xC0,0xC0,0x80},
    {0x20,0x80,0x20},{0xC0,0x40,0xA0},{0xA0,0xA0,0xA0},{0xE0,0xE0,0xE0},
  };

  FILE *F;
  int *T,I,J,K;
  byte *P;
  word A;

  /*** STARTUP CODE starts here: ***/

  T=(int *)"\01\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0";
#ifdef LSB_FIRST
  if(*T!=1)
  {
    printf("********** This machine is high-endian. **********\n");
    printf("Take #define LSB_FIRST out and compile fMSX again.\n");
    return(0);
  }
#else
  if(*T==1)
  {
    printf("********* This machine is low-endian. **********\n");
    printf("Insert #define LSB_FIRST and compile fMSX again.\n");
    return(0);
  }
#endif

  /* Calculate IPeriod from VPeriod/HPeriod */
  if(UPeriod<1) UPeriod=1;
  HPeriod=VPeriod/256;
  CPU.TrapBadOps=Verbose&0x10;
  CPU.IPeriod=HPeriod;
  CPU.IAutoReset=0;

  /* Number of VRAM pages should be a power of 2 */
  for(J=1;J<VRAMPages;J<<=1);
  VRAMPages=J;

  /* Joystick types are in 0..3 range */
  JoyTypeA&=3;
  JoyTypeB&=3;

  /* Check parameters for validity */
  if(MSXVersion>2) MSXVersion=2;
  if(ROMTypeA>5) ROMTypeA=0;
  if(ROMTypeB>5) ROMTypeB=0;
  if((RAMPages<(MSXVersion? 8:4))||(RAMPages>256))
    RAMPages=MSXVersion? 8:4;
  if((VRAMPages<(MSXVersion? 8:2))||(VRAMPages>8))
    VRAMPages=MSXVersion? 8:2;

  /* Zero everyting */
  CasStream=SndStream=PrnStream=ComIStream=ComOStream=NULL;
  FontBuf=NULL;VRAM=NULL;Kanji=NULL;ExitNow=0;
  Drives[0]=Drives[1]=-1;
  CCount=0;

  /* Calculate RAMMask=(2^RAMPages)-1 */
  for(J=1;J<RAMPages;J<<=1);
  RAMMask=J-1;

  /* Initialize ROMMasks to zeroes for now */
  ROMMask[0]=ROMMask[1]=0;

  /* Allocate 16kB for the empty space (scratch RAM) */
  if(Verbose) printf("Allocating 16kB for empty space...");
  if(!(EmptyRAM=malloc(0x4000))) { PRINTFAILED;return(0); }
  memset(EmptyRAM,NORAM,0x4000);
  Chunks[CCount++]=EmptyRAM;

  /* Reset memory map to the empty space */
  for(I=0;I<4;I++)
    for(J=0;J<4;J++)
      for(K=0;K<8;K++)
        MemMap[I][J][K]=EmptyRAM;

  /* Allocate VRAMPages*16kB for VRAM */
  if(Verbose) printf("OK\nAllocating %dkB for VRAM...",VRAMPages*16);
  if(!(VRAM=malloc(VRAMPages*0x4000))) { PRINTFAILED;return(0); }
  memset(VRAM,0x00,VRAMPages*0x4000);
  Chunks[CCount++]=VRAM;

  /* Allocate RAMPages*16kB for RAM */
  if(Verbose) printf("OK\nAllocating %dx16kB RAM pages...",RAMPages);
  if(!(RAMMap[0]=malloc(RAMPages*0x4000))) { PRINTFAILED;return(0); }
  memset(RAMMap[0],NORAM,RAMPages*0x4000);
  Chunks[CCount++]=RAMMap[0];

  /* Set RAM pages */
  for(J=1;J<256;J++) RAMMap[J]=J<RAMPages? RAMMap[0]+J*0x4000:EmptyRAM;

  if(Verbose) printf("OK\nLoading %s ROMs:\n",Versions[MSXVersion]);

  /* Open/load system ROM(s) */
  switch(MSXVersion)
  {
    case 0:
      if(Verbose) printf("  Opening MSX.ROM...");
      P=LoadROM("MSX.ROM",0x8000,0);
      PRINTRESULT(P);
      if(!P) return(0);
      MemMap[0][0][0]=P;
      MemMap[0][0][1]=P+0x2000;
      MemMap[0][0][2]=P+0x4000;
      MemMap[0][0][3]=P+0x6000;
      break;

    case 1:
      if(Verbose) printf("  Opening MSX2.ROM...");
      P=LoadROM("MSX2.ROM",0x8000,0);
      PRINTRESULT(P);
      if(!P) return(0);
      MemMap[0][0][0]=P;
      MemMap[0][0][1]=P+0x2000;
      MemMap[0][0][2]=P+0x4000;
      MemMap[0][0][3]=P+0x6000;
      if(Verbose) printf("  Opening MSX2EXT.ROM...");
      P=LoadROM("MSX2EXT.ROM",0x4000,0);
      PRINTRESULT(P);
      if(!P) return(0);
      MemMap[3][1][0]=P;
      MemMap[3][1][1]=P+0x2000;
      break;

    case 2:
      if(Verbose) printf("  Opening MSX2P.ROM...");
      P=LoadROM("MSX2P.ROM",0x8000,0);
      PRINTRESULT(P);
      if(!P) return(0);
      MemMap[0][0][0]=P;
      MemMap[0][0][1]=P+0x2000;
      MemMap[0][0][2]=P+0x4000;
      MemMap[0][0][3]=P+0x6000;
      if(Verbose) printf("  Opening MSX2PEXT.ROM...");
      P=LoadROM("MSX2PEXT.ROM",0x4000,0);
      PRINTRESULT(P);
      if(!P) return(0);
      MemMap[3][1][0]=P;
      MemMap[3][1][1]=P+0x2000;
      break;
  }

  /* Try loading DiskROM */
  if(P=LoadROM("DISK.ROM",0x4000,0))
  {
    if(Verbose) puts("  Opening DISK.ROM...OK");
    MemMap[3][1][2]=P;
    MemMap[3][1][3]=P+0x2000;
  }

  /* Apply patches to BIOS */
  if(Verbose) printf("  Patching BIOS: ");
  for(J=0;BIOSPatches[J];J++)
  {
    if(Verbose) printf("%04X..",BIOSPatches[J]);
    P=MemMap[0][0][0]+BIOSPatches[J];
    P[0]=0xED;P[1]=0xFE;P[2]=0xC9;
  }
  if(Verbose) printf(".OK\n");

  /* Apply patches to BDOS */
  if(MemMap[3][1][2]!=EmptyRAM)
  {
    if(Verbose) printf("  Patching BDOS: ");
    for(J=0;DiskPatches[J];J++)
    {
      if(Verbose) printf("%04X..",DiskPatches[J]);
      P=MemMap[3][1][2]+DiskPatches[J]-0x4000;
      P[0]=0xED;P[1]=0xFE;P[2]=0xC9;
    }
    if(Verbose) printf(".OK\n");
  }

  /* Try loading cartridges */
  LoadCart(CartA,0);
  LoadCart(CartB,1);

  /* Find an empty slot */
  if(MemMap[1][0][2]==EmptyRAM) J=1;
  else if(MemMap[2][0][2]==EmptyRAM) J=2;
       else J=0;

  /* Try loading MSXDOS2 cartridge, if there */
  /* is an empty slot and DiskROM present    */
  if(J&&(MSXVersion>0)&&(MemMap[3][1][2]!=EmptyRAM))
  {
    if(J==2) ROMTypeB=1; else ROMTypeA=1;
    LoadCart("MSXDOS2.ROM",J-1);
  }

  if(Verbose) printf("Loading other ROMs:\n ");

  /* Try loading CMOS memory contents */
  if(LoadROM("CMOS.ROM",sizeof(RTC),(byte *)RTC))
  { if(Verbose) printf(" CMOS.ROM"); }
  else memcpy(RTC,RTCInit,sizeof(RTC));

  /* Try loading Kanji alphabet ROM */
  if(Kanji=LoadROM("KANJI.ROM",0x20000,0))
  { if(Verbose) printf(" KANJI.ROM"); }

  /* Try loading RS232 support ROM */
  if(P=LoadROM("RS232.ROM",0x4000,0))
  {
    if(Verbose) printf(" RS232.ROM");
    MemMap[3][0][2]=P;
    MemMap[3][0][3]=P+0x2000;
  }

  /* Try loading FM-PAC support ROM */
  if(P=LoadROM("FMPAC.ROM",0x4000,0))
  {
    if(Verbose) printf(" FMPAC.ROM");
    MemMap[3][3][2]=P;
    MemMap[3][3][3]=P+0x2000;
  }

  /* MSX2-dependent ROMs */
  if(MSXVersion>0)
  {
    /* Find an empty slot */
    if(MemMap[1][0][2]==EmptyRAM) J=1;
    else if(MemMap[2][0][2]==EmptyRAM) J=2;
         else J=0;

    /* Try loading PAINTER ROM if slot found */
    if(J&&(P=LoadROM("PAINTER.ROM",0x10000,0)))
    {
      if(Verbose) printf(" PAINTER.ROM");
      for(I=0;I<8;I++) MemMap[J][0][I]=P+I*0x2000;
    }
  }

  /* Try loading font */
  if(Verbose) printf("\nLoading %s font...",FontName);
  FontBuf=LoadROM(FontName,0x800,0);
  PRINTRESULT(FontBuf);

  /* Open stream for a printer */
  if(!PrnName) PrnStream=stdout;
  else
  {
    if(Verbose) printf("Redirecting printer output to %s...",PrnName);
    if(!(PrnStream=fopen(PrnName,"wb"))) PrnStream=stdout;
    PRINTRESULT(PrnStream!=stdout);
  }

  /* Open streams for serial IO */
  if(!ComName) { ComIStream=stdin;ComOStream=stdout; }
  else
  {
    if(Verbose) printf("Redirecting serial I/O to %s...",ComName);
    if(!(ComOStream=ComIStream=fopen(ComName,"r+b")))
    { ComIStream=stdin;ComOStream=stdout; }
    PRINTRESULT(ComOStream!=stdout);
  }

  /* Open soundtrack logfile */
  if(SndName)
  {
    if(Verbose) printf("Logging soundtrack to %s...",SndName);
    SndStream=fopen(SndName,"wb");
    PRINTRESULT(SndStream);
    if(SndStream)
      fwrite("PSG\032\020\074\0\0\0\0\0\0\0\0\0\0",1,16,SndStream);
  }

  /* Open disk images */
  if(ChangeDisk(0,DiskA))
    if(Verbose) printf("Using %s as drive A\n",DiskA);  
  if(ChangeDisk(1,DiskB))
    if(Verbose) printf("Using %s as drive B\n",DiskB);

  /* Open casette image */
  if(CasName)
    if(CasStream=fopen(CasName,"r+b"))
      if(Verbose) printf("Using %s as a tape\n",CasName);

  if(Verbose)
  {
    printf("Attaching %s to joystick port A\n",JoyTypes[JoyTypeA]);
    printf("Attaching %s to joystick port B\n",JoyTypes[JoyTypeB]);
    printf("Initializing memory mappers...");
  }

  for(PSLReg=SSLReg=0x00,J=0;J<4;J++)
  {
    EnWrite[J]=0;                      /* Write protect ON for all slots */
    PSL[J]=SSL[J]=0;                   /* PSL=0:0:0:0, SSL=0:0:0:0       */
    MemMap[3][2][J*2]=RAMMap[3-J];     /* RAMMap=3:2:1:0                 */
    MemMap[3][2][J*2+1]=RAMMap[3-J]+0x2000;
    RAMMapper[J]=3-J;
    RAM[J*2]=MemMap[0][0][J*2];        /* Setting RAM                    */
    RAM[J*2+1]=MemMap[0][0][J*2+1];
  }

  if(Verbose) printf("OK\nInitializing VDP, PSG, and CPU...");

  /* Initialize palette */
  for(J=0;J<16;J++)
    SetColor(J,PalInit[J][0],PalInit[J][1],PalInit[J][2]);

  /* Reset mouse coordinates/counters */
  for(J=0;J<2;J++)
    Buttons[J]=MouseDX[J]=MouseDY[J]=OldMouseX[J]=OldMouseY[J]=MCount[J]=0;

  memcpy(VDP,VDPInit,sizeof(VDP));
  memcpy(VDPStatus,VDPSInit,sizeof(VDPStatus));
  memcpy(PSG,PSGInit,sizeof(PSG));
  memset(KeyMap,0xFF,16);               /* Keyboard         */
  OPLLReg=PSGReg=0;                     /* Sound registers  */
  RTCReg=RTCMode=0;                     /* Clock registers  */
  KanCount=0;KanLetter=0;               /* Kanji extension  */
  ChrTab=ColTab=ChrGen=VRAM;            /* VDP tables       */
  SprTab=SprGen=SprCol=VRAM;
  VPAGE=VRAM;                           /* VRAM page        */
  FGColor=BGColor=XFGColor=XBGColor=0;  /* VDP colors       */
  ScrMode=0;                            /* Screen mode      */
  VKey=PKey=1;WKey=0;                   /* VDP keys         */
  VAddr=0x0000;                         /* VRAM access addr */
  ScanLine=0;                           /* Current scanline */
  VDPData=NORAM;                        /* VDP data buffer  */
  UseFont=0;                            /* Extern. font use */   

  /* Reset CPU and start execution of the code */
  if(Verbose)
  {
    printf("OK\n  %d CPU cycles per HBlank\n",HPeriod);
    printf("  %d CPU cycles per VBlank\n",VPeriod);
    printf("  256 scanlines\nRUNNING ROM CODE...\n");
  }
  ResetZ80(&CPU);
  A=RunZ80(&CPU);

  /* Exiting emulation... */
  if(Verbose) printf("EXITED at PC = %04Xh.\n",A);
  return(1);
}

/** TrashMSX() ***********************************************/
/** Free resources allocated by StartMSX().                 **/
/*************************************************************/
void TrashMSX(void)
{
  int I,J,K;
  FILE *F;
  
  /* Free alocated memory */
  for(J=0;J<CCount;J++) free(Chunks[J]);

  /* Save CMOS RAM, if present */
  if(SaveCMOS)
  {
    if(Verbose) printf("Writing CMOS.ROM...");
    if(F=fopen("CMOS.ROM","wb"))
      if(fwrite(RTC,1,sizeof(RTC),F)!=sizeof(RTC))
      { fclose(F);F=0; }
    if(F) { fclose(F);PRINTOK; }
    else PRINTFAILED;
  }

  /* Close all IO streams */
  if(PrnStream&&(PrnStream!=stdout))   fclose(PrnStream);
  if(ComOStream&&(ComOStream!=stdout)) fclose(ComOStream);
  if(ComIStream&&(ComIStream!=stdin))  fclose(ComIStream);
  if(CasStream) fclose(CasStream);
  if(SndStream) fclose(SndStream);

  /* Close disk images, if present */
  if(Drives[0]>=0) close(Drives[0]);
  if(Drives[1]>=0) close(Drives[1]);
}

/** RdZ80() **************************************************/
/** Z80 emulation calls this function to read a byte from   **/
/** address A of Z80 address space. Now moved to Z80.c and  **/
/** made inlined to speed things up.                        **/
/*************************************************************/
#ifndef FMSX
byte RdZ80(word A)
{
  if(A!=0xFFFF) return(RAM[A>>13][A&0x1FFF]);
  else return(PSL[3]==3? ~SSLReg:RAM[7][0x1FFF]);
}
#endif

/** WrZ80() **************************************************/
/** Z80 emulation calls this function to write byte V to    **/
/** address A of Z80 address space.                         **/
/*************************************************************/
void WrZ80(word A,byte V)
{
  if(A!=0xFFFF)
  {
    if(EnWrite[A>>14]) RAM[A>>13][A&0x1FFF]=V;
    else if((A>0x3FFF)&&(A<0xC000)) MapROM(A,V);
  }
  else
  {
    if(PSL[3]==3) SSlot(V);
    else if(EnWrite[3]) RAM[7][A&0x1FFF]=V;
  }
}

/** InZ80() **************************************************/
/** Z80 emulation calls this function to read a byte from   **/
/** a given I/O port.                                       **/
/*************************************************************/
byte InZ80(word Port)
{
  register byte *P;

  /* MSX only uses 256 IO ports */
  Port&=0xFF;

  /* Return an appropriate port value */
  switch(Port)
  {

case 0x81: return(0x85);                   /* 8251 ACIA status     */
case 0x90: return(0xFD);                   /* Printer READY signal */
case 0xA8: return(PSLReg);                 /* Primary slot state   */
case 0xA9: return(KeyMap[IOReg&0x0F]);     /* Keyboard port        */
case 0xAA: return(IOReg);                  /* General IO register  */
case 0xAB: return(PPIReg);                 /* PPI control register */
case 0xB5: return(RTCIn(RTCReg));          /* RTC registers        */

case 0xFC: /* Mapper page at 0000h */
case 0xFD: /* Mapper page at 4000h */
case 0xFE: /* Mapper page at 8000h */
case 0xFF: /* Mapper page at C000h */
  return(RAMMapper[Port-0xFC]|~RAMMask);

case 0xD9: /* Kanji support */
  Port=Kanji? Kanji[KanLetter+KanCount]:NORAM;
  KanCount=(KanCount+1)&0x1F;
  return(Port);

case 0x80: /* 8251 ACIA data */
  Port=(0x20<<((COMReg&0x0C)>>2))-1;
  return(NORAM&Port);

case 0x98: /* VRAM read port */
  Port=VDPData;
  VDPData=VPAGE[VAddr];
  VAddr=(VAddr+1)&0x3FFF;
  if(!VAddr&&(ScrMode>3))
  {
    VDP[14]=(VDP[14]+1)&(VRAMPages-1);
    VPAGE=VRAM+((long)VDP[14]<<14);
  }
  return(Port);

case 0x99: /* VDP status registers */
  /* Read an appropriate status register */
  Port=VDPStatus[VDP[15]];
  /* Reset interrupt request, if pending */
  CPU.IRequest=INT_NONE;
  /* Reset VAddr latch sequencer */
  VKey=1;
  /* Update status register's contents */
  switch(VDP[15])
  {
    case 0: VDPStatus[0]&=0x5F;break;
    case 1: VDPStatus[1]&=0xFE;break;
    case 2: 
      VDPStatus[2]^=0x60;
      /* VDP Wait */
      if(VDPWait())
      {
        Port |=1;       /* set CE bit */
      }
      else
      {
        Port&=0xfe;     /* reset CE bit */
      }
      break;
    case 7: VDPStatus[7]=VDP[44]=VDPRead();break;
  }
  /* Return the status register value */
  return(Port);

case 0xA2: /* PSG input port */
  /* PSG[14] returns joystick/mouse data */
  if(PSGReg==14)
  {
    int DX,DY,L;

    /* Number of a joystick port */
    Port=PSG[15]&0x40? 1:0;
    L=Port? JoyTypeB:JoyTypeA;

    /* If no joystick, return dummy value */
    if(!L) return(0x7F);

    /* @@@ For debugging purposes */
    /*printf("Reading from PSG[14]: MCount[%d]=%d PSG[15]=%02Xh Value=",Port,MCount[Port],PSG[15]);*/

    /* Poll mouse position, if needed */
    if((L==2)||(MCount[Port]==1))
    {
      /* Read new mouse coordinates */
      L=Mouse(Port);
      Buttons[Port]=(~L>>12)&0x30;
      DY=(L>>8)&0xFF;
      DX=L&0xFF;

      /* Compute offsets and store coordinates  */
      L=OldMouseX[Port]-DX;OldMouseX[Port]=DX;DX=L;
      L=OldMouseY[Port]-DY;OldMouseY[Port]=DY;DY=L;

      /* Adjust offsets */
      MouseDX[Port]=(DX>127? 127:DX<-127? -127:DX)&0xFF;
      MouseDY[Port]=(DY>127? 127:DY<-127? -127:DY)&0xFF;
    }

    /* Determine return value */
    switch(MCount[Port])
    {
      case 0: /* Normal joystick */
        if(PSG[15]&(Port? 0x20:0x10)) Port=0x3F;
        else
          if((Port? JoyTypeB:JoyTypeA)<2) Port=~Joystick(Port)&0x3F;
          else Port=Buttons[Port]
                   |(MouseDX[Port]? (MouseDX[Port]<128? 0x08:0x04):0x0C)
                   |(MouseDY[Port]? (MouseDY[Port]<128? 0x02:0x01):0x03);
        break;

      case 1: Port=(MouseDX[Port]>>4)|Buttons[Port];break;
      case 2: Port=(MouseDX[Port]&0x0F)|Buttons[Port];break;
      case 3: Port=(MouseDY[Port]>>4)|Buttons[Port];break;
      case 4: Port=(MouseDY[Port]&0x0F)|Buttons[Port];break;
    }

    /* @@@ For debugging purposes */
    /*printf("%02Xh\n",Port|0x40);*/

    /* 6th bit is always 1 */
    return(Port|0x40);
  }

  /* PSG[15] resets mouse counters */
  if(PSGReg==15)
  {
    /* @@@ For debugging purposes */
    /*printf("Reading from PSG[15]\n");*/

    /*MCount[0]=MCount[1]=0;*/
    return(PSG[15]&0xF0);
  }

  /* Return PSG[0-13] as they are */
  return(PSG[PSGReg]);

}

  /* Return NORAM for non-existing ports */
  return(NORAM);
}

/** OutZ80() *************************************************/
/** Z80 emulation calls this function to write byte V to a  **/
/** given I/O port.                                         **/
/*************************************************************/
void OutZ80(word Port,byte Value)
{
  static byte PLatch,ALatch;  /* Palette and address buffers */
  register byte I,J,K,*P;  

  Port&=0xFF;
  switch(Port)
  {

case 0x7C: OPLLReg=Value;return;                  /* OPLL Register# */
case 0x81: COMReg=Value;return;                   /* SIO Register#  */
case 0x91: Printer(Value);return;                 /* Printer Data   */
case 0xA0: PSGReg=Value&0x0F;return;              /* PSG Register#  */
case 0xAA: IOReg=Value;PPIOut(IOReg&0x80);return; /* Beeper Bit     */
case 0xB4: RTCReg=Value&0x0F;return;              /* RTC Register#  */ 

case 0xD8: /* Upper bits of Kanji ROM address */
  KanLetter=(KanLetter&0x1F800)|((int)(Value&0x3F)<<5);
  KanCount=0;
  return;

case 0xD9: /* Lower bits of Kanji ROM address */
  KanLetter=(KanLetter&0x007E0)|((int)(Value&0x3F)<<11);
  KanCount=0;
  return;

case 0x80: /* SIO Data */
  J=(0x20<<((COMReg&0x0C)>>2))-1;
  SIOOut(Value&J);
  return;

case 0x7D: /* OPLL Data */
  if(SndStream&&(OPLLReg<0x40))
  { fputc(OPLLReg+0xA0,SndStream);fputc(Value,SndStream); }
  OPLL[OPLLReg]=Value;
  OPLLOut(OPLLReg,Value);
  return;

case 0x98: /* VDP Data */
  if(VKey)
  {
    if(WKey)
    {
      /* VDP set for writing */
      VDPData=VPAGE[VAddr]=Value;
      VAddr=(VAddr+1)&0x3FFF;
    }
    else
    {
      /* VDP set for reading */
      VDPData=VPAGE[VAddr];
      VAddr=(VAddr+1)&0x3FFF;
      VPAGE[VAddr]=Value;
    }

    /* If VAddr rolled over, modify VRAM page# */
    if(!VAddr&&(ScrMode>3)) 
    {
      VDP[14]=(VDP[14]+1)&(VRAMPages-1);
      VPAGE=VRAM+((long)VDP[14]<<14);
    }
  }
  return;

case 0x99: /* VDP Address Latch */
  if(VKey) { ALatch=Value;VKey=0; }
  else
  {
    VKey=1;
    switch(Value&0xC0)
    {
      case 0x80:
        /* Writing into VDP registers */
        VDPOut(Value&0x3F,ALatch);
        break;
      case 0x00:
      case 0x40:
        /* Set the VRAM access address */
        VAddr=(((word)Value<<8)+ALatch)&0x3FFF;
        /* WKey=1 when VDP set for writing into VRAM */
        WKey=Value&0x40;
        /* When set for reading, perform first read */
        if(!WKey)
        {
          VDPData=VPAGE[VAddr];
          VAddr=(VAddr+1)&0x3FFF;
          if(!VAddr&&(ScrMode>3))
          {
            VDP[14]=(VDP[14]+1)&(VRAMPages-1);
            VPAGE=VRAM+((long)VDP[14]<<14);
          }
        }
        break;
    }
  }
  return;

case 0x9A: /* VDP Palette Latch */
  if(PKey) { PLatch=Value;PKey=0; }
  else
  {
    PKey=1;J=VDP[16];
    /* Set new color for palette entry J */
    SetColor(J,(PLatch&0x70)<<1,(Value&0x07)<<5,(PLatch&0x07)<<5);
    /* Next palette entry */
    VDP[16]=(J+1)&0x0F;
  }
  return;

case 0x9B: /* VDP Register Access */
  J=VDP[17]&0x3F;
  if(J!=17) VDPOut(J,Value);
  if(!(VDP[17]&0x80)) VDP[17]=(J+1)&0x3F;
  return;

case 0xA1: /* PSG Data */
  /* PSG[0-13] are sound registers */
  if(PSGReg<14)
  {
    /* Log soundtrack into a file */
    if(SndStream) { fputc(PSGReg,SndStream);fputc(Value,SndStream); }
    /* Update PSG register */
    PSG[PSGReg]=Value;
    /* Play actual sound */
    PSGOut(PSGReg,Value);
    return;
  }

  /* PSG[15] is responsible for joystick/mouse */
  if(PSGReg==15)
  {
    /* @@@ For debugging purposes */
    /*printf("Writing PSG[15] <- %02Xh\n",Value);*/

    /* For mouse, update nibble counter      */
    /* For joystick, set nibble counter to 0 */
    if((Value&0x0C)==0x0C) MCount[1]=0;
    else if((JoyTypeB>2)&&((Value^PSG[15])&0x20))
           MCount[1]+=MCount[1]==4? -3:1;

    /* For mouse, update nibble counter      */
    /* For joystick, set nibble counter to 0 */
    if((Value&0x03)==0x03) MCount[0]=0;
    else if((JoyTypeA>2)&&((Value^PSG[15])&0x10))
           MCount[0]+=MCount[0]==4? -3:1;
  }

  /* Put value into a register */
  PSG[PSGReg]=Value;
  return;

case 0xAB: /* PPI Individual Bit Access */
  if(Value&0x80) PPIReg=Value;
  else
    if(Value&0x01) IOReg|=1<<(7&(Value>>1));
    else           IOReg&=~(1<<(7&(Value>>1)));
  PPIOut(IOReg&0x80);
  return;

case 0xB5: /* RTC Data */
  if(RTCReg<13)
  {
    /* J = register bank# now */
    J=RTCMode&0x03;
    /* Store the value */
    RTC[J][RTCReg]=Value;
    /* If CMOS modified, we need to save it */
    if(J>1) SaveCMOS=1;
    return;
  }
  /* RTC[13] is a register bank# */
  if(RTCReg==13) RTCMode=Value;
  return;

case 0xA8: /* Primary Slot Selector */
  if(Value!=PSLReg)
    for(J=0,PSLReg=Value;J<4;J++,Value>>=2)
    {
      PSL[J]=Value&3;I=J<<1;
      K=PSL[J]==3? SSL[J]:0;
      EnWrite[J]=(K==2)&&(MemMap[3][2][I]!=EmptyRAM);
      RAM[I]=MemMap[PSL[J]][K][I];
      RAM[I+1]=MemMap[PSL[J]][K][I+1];
    }  
  return;

case 0xFC: /* Mapper page at 0000h */
case 0xFD: /* Mapper page at 4000h */
case 0xFE: /* Mapper page at 8000h */
case 0xFF: /* Mapper page at C000h */
  J=Port-0xFC;
  Value&=RAMMask;
  if(RAMMapper[J]!=Value)
  {
    if(Verbose&0x08) printf("RAM-MAPPER: block %d at %Xh\n",Value,J*0x4000);
    I=J<<1;
    RAMMapper[J]=Value;
    MemMap[3][2][I]=RAMMap[Value];
    MemMap[3][2][I+1]=RAMMap[Value]+0x2000;
    if((PSL[J]==3)&&(SSL[J]==2))
    {
      EnWrite[J]=(RAMMap[Value]!=EmptyRAM);
      RAM[I]=MemMap[3][2][I];
      RAM[I+1]=MemMap[3][2][I+1];
    }
  }
  return;

  }
}

/** MapROM() *************************************************/
/** Switch ROM Mapper pages. This function is supposed to   **/
/** be called when ROM page registers are written to.       **/
/*************************************************************/
void MapROM(register word Addr,register byte Value)
{
  byte I,J,K;

  /* J contains page number 0-3  */ 
  J=Addr>>14;

  /* I contains slot number 0/1  */
  if(PSL[J]==1) I=0;
  else if(PSL[J]==2) I=1;
       else return;

  /* K contains MegaROM type 0-5 */
  K=I? ROMTypeB:ROMTypeA;

  /* SCC: types 0, 2, or no cart */
  if((Addr&0xFF00)==0x9800)
    if(!ROMMap[I][0]||(K==0)||(K==2))
    {
      J=Addr&0x00FF;
      if(SndStream&&(J<0x90))
      { fputc(J+0x10,SndStream);fputc(Value,SndStream); }
      SCC[J]=Value;SCCOut(J,Value);
      return;
    }

  /* If no cartridge, exit */
  if(!ROMMap[I][0]) return;

  switch(K)
  {
    case 0: /*** KONAMI 8kB cartridges [general] ***/
      if((Addr&0xFF00)==0x9800)
      {
        J=Addr&0x00FF;
        if(SndStream&&(J<0x90))
        { fputc(J+0x10,SndStream);fputc(Value,SndStream); }
        SCC[J]=Value;SCCOut(J,Value);
        return;
      }
      J=(Addr-0x4000)>>13;K=0;
      break;
    case 1: /*** MSXDOS2 16kB cartridge ***/
      if((Addr<0x6000)||(Addr>0x7FFF)) return;
      J=0;K=1;
      break;
    case 2: /*** KONAMI5 8kB cartridges ***/
      if((Addr&0xFF00)==0x9800)  
      {
        J=Addr&0x00FF;
        if(SndStream&&(J<0x90))
        { fputc(J+0x10,SndStream);fputc(Value,SndStream); }
        SCC[J]=Value;SCCOut(J,Value);
        return;
      }
      Addr-=0x5000;if(Addr&0x9FFF) return;
      J=Addr>>13;K=0;
      break;
    case 3: /*** KONAMI4 8kB cartridges ***/
      Addr-=0x4000;if(Addr&0x9FFF) return;
      J=Addr>>13;K=0;
      break;
    case 4: /*** ASCII 8kB cartridges ***/
      if((Addr>=0x8000)||(Addr<0x6000)) return;
      J=(Addr&0x1800)>>11;K=0;
      break;
    case 5: /*** ASCII 16kB cartridges ***/
      if((Addr>=0x8000)||(Addr<0x6000)) return;
      J=(Addr&0x1000)>>11;K=1;
      break;
    default:
      printf("Unknown MegaROM type %d\n",I? ROMTypeB:ROMTypeA);
      return; 
  }

  Value&=ROMMask[I];
  if(K) Value<<=1;

  if(Value!=ROMMapper[I][J])
  {
    ROMMapper[I][J]=Value;
    RAM[J+2]=MemMap[I+1][0][J+2]=ROMMap[I][Value];
    if(K)
    {
      if(Verbose&0x08)
        printf("ROM-MAPPER: 16kB block %d at %04Xh\n",Value,J*0x2000+0x4000);
      ROMMapper[I][J+1]=Value+1;
      RAM[J+3]=MemMap[I+1][0][J+3]=ROMMap[I][Value+1];
    }
    else
      if(Verbose&0x08)
        printf("ROM-MAPPER: 8kB block %d at %04Xh\n",Value,J*0x2000+0x4000);
  }
}

/** SSlot() **************************************************/
/** Switch secondary memory slots. This function is called  **/
/** when value in (FFFFh) changes.                          **/
/*************************************************************/
void SSlot(register byte Value)
{
  register byte I,J,K,*P;
  
  if(SSLReg!=Value)
  {
    SSLReg=Value;

    for(J=0;J<4;J++,Value>>=2)
    {
      SSL[J]=Value&3;
      if(PSL[J]==3)
      {
        I=J<<1;
        EnWrite[J]=(SSL[J]==2)&&(MemMap[3][2][I]!=EmptyRAM);
        RAM[I]=MemMap[3][SSL[J]][I];
        RAM[I+1]=MemMap[3][SSL[J]][I+1];
      }
    }
  }
}

#ifdef ZLIB
#define fopen          gzopen
#define fclose         gzclose
#define fread(B,N,L,F) gzread(F,B,(L)*(N))
#define fseek          gzseek
#define frewind        gzrewind
#define fgetc          gzgetc
#endif

/** LoadROM() ************************************************/
/** Load a file, allocating memory as needed. Returns addr. **/
/** of the alocated space or 0 if failed.                   **/
/*************************************************************/
byte *LoadROM(char *Name,unsigned int Size,byte *Buf)
{
  FILE *F;
  byte *P;

  /* Can't give address without size! */
  if(Buf&&!Size) return(0);

  /* Open file */
  if(!(F=fopen(Name,"rb"))) return(0);

  /* Determine data size, if wasn't given */
  if(!Size)
  {
    if(fseek(F,0,SEEK_END)<0) { fclose(F);return(0); }
    if((Size=ftell(F))<=0)    { fclose(F);return(0); }
    rewind(F);
  }

  /* Allocate memory */
  P=Buf? Buf:malloc(Size);
  if(!P)
  {
    fclose(F);
    return(0);
  }

  /* Read data */
  if(fread(P,1,Size,F)!=Size)
  {
    if(!Buf) free(P);
    fclose(F);
    return(0);
  }

  /* Add a new chunk to free */
  if(!Buf) Chunks[CCount++]=P;

  /* Done */
  fclose(F);
  return(P);
}

/** LoadCart() ***********************************************/
/** Load a cartridge ROM from .ROM file into given slot.    **/
/*************************************************************/
int LoadCart(char *Name,int Slot)
{
  static char *ROMNames[] = 
  { 
    "KONAMI/8kB","MSXDOS2","KONAMI5/8kB",
    "KONAMI4/8kB","ASCII/8kB","ASCII/16kB"
  };
  int C1,C2,C3;
  FILE *F;

  /* Check slot #, try to open file */
  if((Slot!=0)&&(Slot!=1)) return(0);
  if(!(F=fopen(Name,"rb"))) return(0);
  if(Verbose) printf("Found %s:\n",Name);

  /* Check "AB" signature in a file */
  C1=fgetc(F);
  C2=fgetc(F);
  if((C1!='A')||(C2!='B'))
  { 
    if(Verbose) puts("  Not a valid cartridge ROM");
    fclose(F);
    return(0);
  }

  if(Verbose) printf("  Cartridge %c: Loading...",'A'+Slot);

  /* Figure out file length */
  if(fseek(F,0,SEEK_END)<0) { PRINTFAILED;fclose(F);return(0); }
  if((C1=ftell(F))<=0)      { PRINTFAILED;fclose(F);return(0); }
  fclose(F);

  /* Compute size in 8kB pages */
  C1>>=13;

  /* Show ROM type and size */
  if(Verbose)
    printf
    ("%dkB %s ROM...",C1*8,C1>4? ROMNames[Slot? ROMTypeB:ROMTypeA]:"NORMAL");

  /* Calculate 2^n closest to number of pages */
  for(C3=1;C3<C1;C3<<=1);

  /* Try loading ROM */
  ROMMap[Slot][0]=LoadROM(Name,C1*0x2000,0);
  if(!ROMMap[Slot][0]) { PRINTFAILED;return(0); }

  /* Set ROMMap */
  for(C2=1;C2<C3;C2++)
    ROMMap[Slot][C2]=C2<C1? ROMMap[Slot][0]+C2*0x2000:EmptyRAM;

  /* Set memory map depending on the ROM size */
  switch(C1)
  {
    case 1:
      /* 8kB ROMs are mirrored 8 times: 0:0:0:0:0:0:0:0 */
      MemMap[Slot+1][0][0]=ROMMap[Slot][0];
      MemMap[Slot+1][0][1]=ROMMap[Slot][0];
      MemMap[Slot+1][0][2]=ROMMap[Slot][0];
      MemMap[Slot+1][0][3]=ROMMap[Slot][0];
      MemMap[Slot+1][0][4]=ROMMap[Slot][0];
      MemMap[Slot+1][0][5]=ROMMap[Slot][0];
      MemMap[Slot+1][0][6]=ROMMap[Slot][0];
      MemMap[Slot+1][0][7]=ROMMap[Slot][0];
      break;

    case 2:
      /* 16kB ROMs are mirrored 4 times: 0:1:0:1:0:1:0:1 */
      MemMap[Slot+1][0][0]=ROMMap[Slot][0];
      MemMap[Slot+1][0][1]=ROMMap[Slot][1];
      MemMap[Slot+1][0][2]=ROMMap[Slot][0];
      MemMap[Slot+1][0][3]=ROMMap[Slot][1];
      MemMap[Slot+1][0][4]=ROMMap[Slot][0];
      MemMap[Slot+1][0][5]=ROMMap[Slot][1];
      MemMap[Slot+1][0][6]=ROMMap[Slot][0];
      MemMap[Slot+1][0][7]=ROMMap[Slot][1];
      break;

    case 3:
    case 4:
      /* 24kB and 32kB ROMs are mirrored twice: 0:1:0:1:2:3:2:3 */
      MemMap[Slot+1][0][0]=ROMMap[Slot][0];
      MemMap[Slot+1][0][1]=ROMMap[Slot][1];
      MemMap[Slot+1][0][2]=ROMMap[Slot][0];
      MemMap[Slot+1][0][3]=ROMMap[Slot][1];
      MemMap[Slot+1][0][4]=ROMMap[Slot][2];
      MemMap[Slot+1][0][5]=ROMMap[Slot][3];
      MemMap[Slot+1][0][6]=ROMMap[Slot][2];
      MemMap[Slot+1][0][7]=ROMMap[Slot][3];
      break;

    default:
      /* MegaROMs are switched into 4000h-BFFFh */
      MemMap[Slot+1][0][2]=ROMMap[Slot][0];
      MemMap[Slot+1][0][3]=ROMMap[Slot][1];
      MemMap[Slot+1][0][4]=ROMMap[Slot][2];
      MemMap[Slot+1][0][5]=ROMMap[Slot][3];
      ROMMapper[Slot][0]=0;
      ROMMapper[Slot][1]=1;
      ROMMapper[Slot][2]=2;
      ROMMapper[Slot][3]=3;
      break;
  }

  /* Show starting address */
  if(Verbose)
    printf("starts at %04Xh...OK\n",ROMMap[Slot][0][2]+256*ROMMap[Slot][0][3]);

  /* Assign ROMMask for MegaROMs */
  ROMMask[Slot]=C1>4? C3-1:0x00;

  /* Done loading cartridge */
  return(1);
}

#ifdef ZLIB
#undef fopen
#undef fclose
#undef fread
#undef fseek
#undef frewind
#undef fgetc
#endif

/** VDPOut() *************************************************/
/** Write value into a given VDP register.                  **/
/*************************************************************/
void VDPOut(register byte R,register byte V)
{ 
  register byte I,J,*P;  
  register pair K;

  switch(R)  
  {
    case  0: switch(((V&0x0E)>>1)|(VDP[1]&0x18))
             {
               case 0x10: J=0;break;
               case 0x00: J=1;break;
               case 0x01: J=2;break;
               case 0x08: J=3;break;
               case 0x02: J=4;break;
               case 0x03: J=5;break;
               case 0x04: J=6;break;
               case 0x05: J=7;break;
               case 0x07: J=8;break;
               case 0x12: J=MAXSCREEN+1;break;
               default:   J=ScrMode;
             }
             if(J!=ScrMode)
             {
               I=((J>6)&&(J!=MAXSCREEN+1))? 11:10;
               ChrTab=VRAM+((long)(VDP[2]&MSK[J].R2)<<I);
               ChrGen=VRAM+((long)(VDP[4]&MSK[J].R4)<<11);
               ColTab=VRAM+((long)(VDP[3]&MSK[J].R3)<<6)
                          +((long)VDP[10]<<14);
	       SprTab=VRAM+((long)(VDP[5]&MSK[J].R5)<<7)
                          +((long)VDP[11]<<15);
               SprGen=VRAM+((long)VDP[6]<<11);
               ScrMode=J;
               /* clear_fmsx_screen(1); */
             }
             break;
    case  1: switch(((VDP[0]&0x0E)>>1)|(V&0x18))
             {
               case 0x10: J=0;break;
               case 0x00: J=1;break;
               case 0x01: J=2;break;
               case 0x08: J=3;break;
               case 0x02: J=4;break;
               case 0x03: J=5;break;
               case 0x04: J=6;break;
               case 0x05: J=7;break;
               case 0x07: J=8;break;
               case 0x12: J=MAXSCREEN+1;break;
               default:   J=ScrMode;
             }   
             if(J!=ScrMode)
             {
               I=((J>6)&&(J!=MAXSCREEN+1))? 11:10;
               ChrTab=VRAM+((long)(VDP[2]&MSK[J].R2)<<I);
               ChrGen=VRAM+((long)(VDP[4]&MSK[J].R4)<<11);
               ColTab=VRAM+((long)(VDP[3]&MSK[J].R3)<<6)
                          +((long)VDP[10]<<14);
               SprTab=VRAM+((long)(VDP[5]&MSK[J].R5)<<7)  
                          +((long)VDP[11]<<15);
               SprGen=VRAM+((long)VDP[6]<<11);
               ScrMode=J;
               /* clear_fmsx_screen(1); */
             }
             break;       
    case  2: J=((ScrMode>6)&&(ScrMode!=MAXSCREEN+1))? 11:10;
             ChrTab=VRAM+((long)(V&MSK[ScrMode].R2)<<J);
             break;
    case  3: ColTab=VRAM+((long)(V&MSK[ScrMode].R3)<<6)+((long)VDP[10]<<14);
             break;
    case  4: ChrGen=VRAM+((long)(V&MSK[ScrMode].R4)<<11);
             break;
    case  5: SprTab=VRAM+((long)(V&MSK[ScrMode].R5)<<7)+((long)VDP[11]<<15);
             break;
    case  6: V&=0x3F;SprGen=VRAM+((long)V<<11);break;
    case  7: FGColor=V>>4;BGColor=V&0x0F;break;
    case 10: V&=0x07;
             ColTab=VRAM+((long)(VDP[3]&MSK[ScrMode].R3)<<6)+((long)V<<14);
             break;
    case 11: V&=0x03;
             SprTab=VRAM+((long)(VDP[5]&MSK[ScrMode].R5)<<7)+((long)V<<15);
             break;
    case 14: V&=VRAMPages-1;VPAGE=VRAM+((long)V<<14);
             break;
    case 15: V&=0x0F;break;
    case 16: V&=0x0F;break;
    case 17: V&=0xBF;break;
    case 44: VDPWrite(V);break;
    case 46: VDPDraw(V);break;
  }

  /* Write value into a register */
  VDP[R]=V;
} 

/** ChangeDisk() *********************************************/
/** Change disk image in a given drive. Closes current disk **/
/** image if Name=0 was given. Returns 1 on success or 0 on **/
/** failure.                                                **/
/*************************************************************/
byte ChangeDisk(byte Drive,char *Name)
{
  /* We only have two drives, A and B */
  if(Drive>1) return(0);
  /* Close previous disk image */
  if(Drives[Drive]>=0) { close(Drives[Drive]);Drives[Drive]=-1; }
  /* If no disk image given, consider drive empty */
  if(!Name) return(1);
  /* Open new disk image */
  Drives[Drive]=open(Name,O_RDWR|O_BINARY);
  /* Return operation result */
  return(Drives[Drive]>=0);
}

/** Printer() ************************************************/
/** Send a character to the printer.                        **/
/*************************************************************/
void Printer(byte V) { fputc(V,PrnStream); }

/** SIOOut() *************************************************/
/** Send a character to the serial port.                    **/
/*************************************************************/
void SIOOut(byte V) { fputc(V,ComOStream); }

/** SIOIn() **************************************************/
/** Receive a character from the serial port.               **/
/*************************************************************/
byte SIOIn(void) { return(NORAM);/*fgetc(ComIStream));*/ }

/** RTCIn() **************************************************/
/** Read value from a given RTC register.                   **/
/*************************************************************/
byte RTCIn(register byte R)
{
  static time_t PrevTime;
  static struct tm TM;
  register byte J;
  time_t CurTime;

  /* Only 16 registers/mode */
  R&=0x0F;

  /* Bank mode 0..3 */
  J=RTCMode&0x03;

  if(R>12) J=R==13? RTCMode:NORAM;
  else
    if(J) J=RTC[J][R];
    else
    {
      /* Retrieve system time if any time passed */
      CurTime=time(NULL);
      if(CurTime!=PrevTime)
      {
        TM=*localtime(&CurTime);
        PrevTime=CurTime;
      }

      /* Parse contents of last retrieved TM */
      switch(R)
      {
        case 0:  J=TM.tm_sec%10;break;
        case 1:  J=TM.tm_sec/10;break;
        case 2:  J=TM.tm_min%10;break;
        case 3:  J=TM.tm_min/10;break;
        case 4:  J=TM.tm_hour%10;break;
        case 5:  J=TM.tm_hour/10;break;
        case 6:  J=TM.tm_wday;break;
        case 7:  J=TM.tm_mday%10;break;
        case 8:  J=TM.tm_mday/10;break;
        case 9:  J=(TM.tm_mon+1)%10;break;
        case 10: J=(TM.tm_mon+1)/10;break;
        case 11: J=(TM.tm_year-80)%10;break;
        case 12: J=((TM.tm_year-80)/10)%10;break;
        default: J=0x0F;break;
      } 
    }

  /* Four upper bits are always high */
  return(J|0xF0);
}

/** LoopZ80() ************************************************/
/** Refresh screen, check keyboard and sprites. Call this   **/
/** function on each interrupt.                             **/
/*************************************************************/
word LoopZ80(Z80 *R)
{
  static byte BFlag=0;
  static byte BCount=0;
  static int UCount=1;
  register byte J;

  /* New line */
  ScanLine++;

  /* increment HSync counter */
  IncTotalHSync();

  /* First scanline of VBlank */
  if(ScanLine==212)
  {
    /* Check keyboard */
    Keyboard();

    /* Check sprites */
    if(!SpritesOFF&&ScrMode&&(ScrMode<MAXSCREEN+1)) CheckSprites();

    /* Refresh sprites and display */
    if(UCount) UCount--; else { UCount=UPeriod-1;RefreshScreen(); }

    /* Set VBlank interrupt bit, reset ??? */
    VDPStatus[0]=(VDPStatus[0]&0xDF)|0x80;

    /* Write VBlank timestamp into sound log */
    if(SndStream) fputc(0xFF,SndStream);

    /* Generate VBlank interrupt */
    if((VDPStatus[0]&0x80)&&(VDP[1]&0x20)) R->IRequest=INT_IRQ;

    /* Exit emulation if requested */
    if(ExitNow) return(INT_QUIT);

    /* Blinking for TEXT80 */
    if(BCount) BCount--;
    else
    {
      BFlag=!BFlag;
      if(!VDP[13]) { XFGColor=FGColor;XBGColor=BGColor; }
      else
      {
        BCount=BFlag? VDP[13]&0x0F:VDP[13]>>4;
        if(BCount)
          if(BFlag) { XFGColor=FGColor;XBGColor=BGColor; }
          else      { XFGColor=VDP[12]>>4;XBGColor=VDP[12]&0x0F; }
      }
    }
  }

  /* First scanline of the screen */
  if(!ScanLine)
  {
    /* No interrupts yet */
    VDPStatus[0]&=0x7F;
    R->IRequest=INT_NONE;

    /* reset HSync counter */
    /* ResetTotalHSync(); */

  }

  /* Set HBlank flag on line coincidence, generate interrupt */
  J=(byte)(ScanLine+VDP[23])-VDP[19];
  if((J==3)&&(VDP[0]&0x10))
  { VDPStatus[1]|=0x01;R->IRequest=INT_IRQ; }

  /* Refresh scanline */
  if(!UCount&&(ScanLine<(ScanLines212? 212:192)))
    (RefreshLine[ScrMode])(ScanLine);

  /* Return whatever interrupt is pending */
  return(VKey? R->IRequest:INT_NONE);
}

/** CheckSprites() *******************************************/
/** Check for sprite collisions and 5th/9th sprite in a     **/
/** row.                                                    **/
/*************************************************************/
void CheckSprites(void)
{
  register word LS,LD;
  register byte DH,DV,*PS,*PD,*T;
  byte I,J,N,M,*S,*D;

  VDPStatus[0]=(VDPStatus[0]&0x9F)|0x1F;
  for(N=0,S=SprTab;(N<32)&&(S[0]!=208);N++,S+=4);
  M=SolidColor0;

  if(Sprites16x16)
  {
    for(J=0,S=SprTab;J<N;J++,S+=4)
      if((S[3]&0x0F)||M)
        for(I=J+1,D=S+4;I<N;I++,D+=4)
          if((D[3]&0x0F)||M) 
          {
            DV=S[0]-D[0];
            if((DV<16)||(DV>240))
	    {
              DH=S[1]-D[1];
              if((DH<16)||(DH>240))
	      {
                PS=SprGen+((long)(S[2]&0xFC)<<3);
                PD=SprGen+((long)(D[2]&0xFC)<<3);
                if(DV<16) PD+=DV; else { DV=256-DV;PS+=DV; }
                if(DH>240) { DH=256-DH;T=PS;PS=PD;PD=T; }
                while(DV<16)
                {
                  LS=((word)*PS<<8)+*(PS+16);
                  LD=((word)*PD<<8)+*(PD+16);
                  if(LD&(LS>>DH)) break;
                  else { DV++;PS++;PD++; }
                }
                if(DV<16) { VDPStatus[0]|=0x20;return; }
              }
            }
          }
  }
  else
  {
    for(J=0,S=SprTab;J<N;J++,S+=4)
      if((S[3]&0x0F)||M)
        for(I=J+1,D=S+4;I<N;I++,D+=4)
          if((D[3]&0x0F)||M)
          {
            DV=S[0]-D[0];
            if((DV<8)||(DV>248))
            {
              DH=S[1]-D[1];
              if((DH<8)||(DH>248))
              {
                PS=SprGen+((long)S[2]<<3);
                PD=SprGen+((long)D[2]<<3);
                if(DV<8) PD+=DV; else { DV=256-DV;PS+=DV; }
                if(DH>248) { DH=256-DH;T=PS;PS=PD;PD=T; }
                while((DV<8)&&!(*PD&(*PS>>DH))) { DV++;PS++;PD++; }
                if(DV<8) { VDPStatus[0]|=0x20;return; }
              }
            }
          }
  }
}
